/**
   UnoAnalogLogger is a modification of SdFat AnalogBinLogger. https://github.com/greiman/SdFat
   modications by Matthew Ford
   (c)2019 Forward Computing and Control Pty. Ltd. NSW Australia
   All rights reserverd.
   This code is not warranted to be fit for any purpose. You may only use it at your own risk.
   This code may be freely used for both private and commercial use
   Provide this copyright is maintained.

*/
/**
   from https://github.com/greiman/SdFat AnalogBinLogger
   This program logs data from the Arduino ADC to a binary file.

   Samples are logged at regular intervals. Each Sample consists of the ADC
   values for the analog pins defined in the PIN_LIST array.  The pins numbers
   may be in any order.

   Edit the configuration constants below to set the sample pins, sample rate,
   and other configuration values.

   If your SD card has a long write latency, it may be necessary to use
   slower sample rates.  Using a Mega Arduino helps overcome latency
   problems since 13 512 byte buffers will be used.

   Each 512 byte data block in the file has a four byte header followed by up
   to 508 bytes of data. (508 values in 8-bit mode or 254 values in 10-bit mode)
   Each block contains an integral number of samples with unused space at the
   end of the block.

   Data is written to the file using a SD multiple block write command.
*/
#ifdef __AVR__
#include <SPI.h>
#include "SdFat.h"
#include "FreeStack.h"
#include "AnalogBinLogger.h"
#include <millisDelay.h>

#define VERSION "13"
//------------------------------------------------------------------------------
// Analog pin number list for a sample.  Pins may be in any order and pin
// numbers may be repeated.
const uint8_t PIN_LIST[] = {0, 1};//, 2, 3, 4};
//------------------------------------------------------------------------------
// Sample rate in samples per second.
const float SAMPLE_RATE = 1000;  // Must be 0.25 or greater.

// The interval between samples in seconds, SAMPLE_INTERVAL, may be set to a
// constant instead of being calculated from SAMPLE_RATE.  SAMPLE_RATE is not
// used in the code below.  For example, setting SAMPLE_INTERVAL = 2.0e-4
// will result in a 200 microsecond sample interval.
const float SAMPLE_INTERVAL = 1.0 / SAMPLE_RATE;

// Setting ROUND_SAMPLE_INTERVAL non-zero will cause the sample interval to
// be rounded to a a multiple of the ADC clock period and will reduce sample
// time jitter.
#define ROUND_SAMPLE_INTERVAL 1
//------------------------------------------------------------------------------
// Reference voltage.  See the processor data-sheet for reference details.
// uint8_t const ADC_REF = 0; // External Reference AREF pin.
uint8_t const ADC_REF = (1 << REFS0);  // Vcc Reference.
// uint8_t const ADC_REF = (1 << REFS1);  // Internal 1.1 (only 644 1284P Mega)
// uint8_t const ADC_REF = (1 << REFS1) | (1 << REFS0);  // Internal 1.1 or 2.56
//------------------------------------------------------------------------------
// File definitions.
//
// Maximum file size in blocks.
// The program creates a contiguous file with FILE_BLOCK_COUNT 512 byte blocks.
// This file is flash erased using special SD commands.  The file will be
// truncated if logging is stopped early.
const uint32_t FILE_BLOCK_COUNT = 256000;

// log file base name.  Must be six characters or less.
#define FILE_BASE_NAME "log"
// usually use FILE_BASE_EXTRA  "00.bin"
#define FILE_BASE_EXTRA  "00.bin"
#define FILE_BASE_META_EXTRA  ".met"
// max number of files to write max is 99 ~12Gig
// set to 3 for testing  normally set to 99
#define FILE_COUNT 99

// Set RECORD_EIGHT_BITS non-zero to record only the high 8-bits of the ADC.
#define RECORD_EIGHT_BITS 0
//------------------------------------------------------------------------------
// Pin definitions.
//
// Digital pin to indicate an error, set to -1 if not used.
// The led blinks for fatal errors.
// The led goes on solid for SD write overrun errors (missed samples) and logging continues.
const int8_t ERROR_LED_PIN = 3;
const int8_t ERROR_LED_GND_PIN = 2; // set this low to supply GND for error led, set to -1 if not used

// SD chip select pin.
const uint8_t SD_CS_PIN = SS;
//------------------------------------------------------------------------------
// Buffer definitions.
//
// The logger will use SdFat's buffer plus BUFFER_BLOCK_COUNT additional
// buffers.  QUEUE_DIM must be a power of two larger than
//(BUFFER_BLOCK_COUNT + 1).
//
#if RAMEND < 0X8FF
#error Too little SRAM
//
#elif RAMEND < 0X10FF
// Use total of two 512 byte buffers.
const uint8_t BUFFER_BLOCK_COUNT = 1;
// Dimension for queues of 512 byte SD blocks.
const uint8_t QUEUE_DIM = 4;  // Must be a power of two!
//
#elif RAMEND < 0X20FF
// Use total of five 512 byte buffers.
const uint8_t BUFFER_BLOCK_COUNT = 4;
// Dimension for queues of 512 byte SD blocks.
const uint8_t QUEUE_DIM = 8;  // Must be a power of two!
//
#elif RAMEND < 0X40FF
// Use total of 13 512 byte buffers.
const uint8_t BUFFER_BLOCK_COUNT = 12;
// Dimension for queues of 512 byte SD blocks.
const uint8_t QUEUE_DIM = 16;  // Must be a power of two!
//
#else  // RAMEND
// Use total of 29 512 byte buffers.
const uint8_t BUFFER_BLOCK_COUNT = 28;
// Dimension for queues of 512 byte SD blocks.
const uint8_t QUEUE_DIM = 32;  // Must be a power of two!
#endif  // RAMEND
//------------------------------------------------------------------------------
// ADC clock rate.
// The ADC clock rate is normally calculated from the pin count and sample
// interval.  The calculation attempts to use the lowest possible ADC clock
// rate.
//
// You can select an ADC clock rate by defining the symbol ADC_PRESCALER to
// one of these values.  You must choose an appropriate ADC clock rate for
// your sample interval.
// #define ADC_PRESCALER 7 // F_CPU/128 125 kHz on an Uno
// #define ADC_PRESCALER 6 // F_CPU/64  250 kHz on an Uno
// #define ADC_PRESCALER 5 // F_CPU/32  500 kHz on an Uno
// #define ADC_PRESCALER 4 // F_CPU/16 1000 kHz on an Uno
// #define ADC_PRESCALER 3 // F_CPU/8  2000 kHz on an Uno (8-bit mode only)
//==============================================================================
// End of configuration constants.
//==============================================================================

// Size of file base name.  Must not be larger than six.
const uint8_t BASE_NAME_SIZE = sizeof(FILE_BASE_NAME) - 1;

// Number of analog pins to log.
const uint8_t PIN_COUNT = sizeof(PIN_LIST) / sizeof(PIN_LIST[0]);

// Minimum ADC clock cycles per sample interval
const uint16_t MIN_ADC_CYCLES = 15;

// Extra cpu cycles to setup ADC with more than one pin per sample.
const uint16_t ISR_SETUP_ADC = PIN_COUNT > 1 ? 100 : 0;

// Maximum cycles for timer0 system interrupt, millis, micros.
const uint16_t ISR_TIMER0 = 160;
//==============================================================================
SdFat sd;

SdBaseFile binFile;

const size_t NAME_DIM = 13;
char binName[NAME_DIM] = FILE_BASE_NAME FILE_BASE_EXTRA;
bool processAllFiles = false; // set to true is e or c to process all the files
bool errorBlink = false; // set to true on some error
millisDelay blinkDelay;
uint32_t lastOverruns = 0;


#if RECORD_EIGHT_BITS
const size_t SAMPLES_PER_BLOCK = DATA_DIM8 / PIN_COUNT;
typedef block8_t block_t;
#else  // RECORD_EIGHT_BITS
const size_t SAMPLES_PER_BLOCK = DATA_DIM16 / PIN_COUNT;
typedef block16_t block_t;
#endif // RECORD_EIGHT_BITS

block_t* emptyQueue[QUEUE_DIM];
uint8_t emptyHead;
uint8_t emptyTail;

block_t* fullQueue[QUEUE_DIM];
volatile uint8_t fullHead;  // volatile insures non-interrupt code sees changes.
uint8_t fullTail;

// queueNext assumes QUEUE_DIM is a power of two
inline uint8_t queueNext(uint8_t ht) {
  return (ht + 1) & (QUEUE_DIM - 1);
}
//==============================================================================
// Interrupt Service Routines

// Pointer to current buffer.
block_t* isrBuf;

// Need new buffer if true.
bool isrBufNeeded = true;

// overrun count, i.e. missed samples
uint16_t isrOver = 0;

// ADC configuration for each pin.
uint8_t adcmux[PIN_COUNT];
uint8_t adcsra[PIN_COUNT];
uint8_t adcsrb[PIN_COUNT];
uint8_t adcindex = 1;

// Insure no timer events are missed.
volatile bool timerError = false;
volatile bool timerFlag = false;
//------------------------------------------------------------------------------
// ADC done interrupt.
ISR(ADC_vect) {
  // Read ADC data.
#if RECORD_EIGHT_BITS
  uint8_t d = ADCH;
#else  // RECORD_EIGHT_BITS
  // This will access ADCL first.
  uint16_t d = ADC;
#endif  // RECORD_EIGHT_BITS

  if (isrBufNeeded && emptyHead == emptyTail) {
    // no buffers - count overrun
    // missed this sample
    if (isrOver < 0XFFFF) {
      isrOver++;
    }

    // Avoid missed timer error.
    timerFlag = false;
    return;
  }
  // Start ADC
  if (PIN_COUNT > 1) {
    ADMUX = adcmux[adcindex];
    ADCSRB = adcsrb[adcindex];
    ADCSRA = adcsra[adcindex];
    if (adcindex == 0) {
      timerFlag = false;
    }
    adcindex =  adcindex < (PIN_COUNT - 1) ? adcindex + 1 : 0;
  } else {
    timerFlag = false;
  }
  // Check for buffer needed.
  if (isrBufNeeded) {
    // Remove buffer from empty queue.
    isrBuf = emptyQueue[emptyTail];
    emptyTail = queueNext(emptyTail);
    isrBuf->count = 0;
    isrBuf->overrun = isrOver;
    isrBufNeeded = false;
  }
  // Store ADC data.
  isrBuf->data[isrBuf->count++] = d;

  // Check for buffer full.
  if (isrBuf->count >= PIN_COUNT * SAMPLES_PER_BLOCK) {
    // Put buffer isrIn full queue.
    uint8_t tmp = fullHead;  // Avoid extra fetch of volatile fullHead.
    fullQueue[tmp] = (block_t*)isrBuf;
    fullHead = queueNext(tmp);

    // Set buffer needed and clear overruns.
    isrBufNeeded = true;
    isrOver = 0;
  }
}
//------------------------------------------------------------------------------
// timer1 interrupt to clear OCF1B
ISR(TIMER1_COMPB_vect) {
  // Make sure ADC ISR responded to timer event.
  if (timerFlag) {
    timerError = true;
  }
  timerFlag = true;
}
//==============================================================================
// Error messages stored in flash.
#define error(msg) {sd.errorPrint(F(msg));fatalBlink();}
//------------------------------------------------------------------------------
//
void fatalBlink() {
  errorBlink = true;
  //  while (true) {
  //    if (ERROR_LED_PIN >= 0) {
  //      digitalWrite(ERROR_LED_PIN, HIGH);
  //      delay(200);
  //      digitalWrite(ERROR_LED_PIN, LOW);
  //      delay(200);
  //    }
  //  }
}
//==============================================================================
#if ADPS0 != 0 || ADPS1 != 1 || ADPS2 != 2
#error unexpected ADC prescaler bits
#endif
//------------------------------------------------------------------------------
// initialize ADC and timer1
void adcInit(metadata_t* meta) {
  uint8_t adps;  // prescaler bits for ADCSRA
  uint32_t ticks = F_CPU * SAMPLE_INTERVAL + 0.5; // Sample interval cpu cycles.

  if (ADC_REF & ~((1 << REFS0) | (1 << REFS1))) {
    error("Invalid ADC reference");
    while (1) {
      ; // stop here
    }
  }
#ifdef ADC_PRESCALER
  if (ADC_PRESCALER > 7 || ADC_PRESCALER < 2) {
    error("Invalid ADC prescaler");
    while (1) {
      ; // stop here
    }

  }
  adps = ADC_PRESCALER;
#else  // ADC_PRESCALER
  // Allow extra cpu cycles to change ADC settings if more than one pin.
  int32_t adcCycles = (ticks - ISR_TIMER0) / PIN_COUNT - ISR_SETUP_ADC;

  for (adps = 7; adps > 0; adps--) {
    if (adcCycles >= (MIN_ADC_CYCLES << adps)) {
      break;
    }
  }
#endif  // ADC_PRESCALER
  meta->adcFrequency = F_CPU >> adps;
  if (meta->adcFrequency > (RECORD_EIGHT_BITS ? 2000000 : 1000000)) {
    error("Sample Rate Too High");
    while (1) {
      ; // stop here
    }

  }
#if ROUND_SAMPLE_INTERVAL
  // Round so interval is multiple of ADC clock.
  ticks += 1 << (adps - 1);
  ticks >>= adps;
  ticks <<= adps;
#endif  // ROUND_SAMPLE_INTERVAL

  if (PIN_COUNT > sizeof(meta->pinNumber) / sizeof(meta->pinNumber[0])) {
    error("Too many pins");
    while (1) {
      ; // stop here
    }

  }
  meta->pinCount = PIN_COUNT;
  meta->recordEightBits = RECORD_EIGHT_BITS;

  for (int i = 0; i < PIN_COUNT; i++) {
    uint8_t pin = PIN_LIST[i];
    if (pin >= NUM_ANALOG_INPUTS) {
      error("Invalid Analog pin number");
      while (1) {
        ; // stop here
      }

    }
    meta->pinNumber[i] = pin;

    // Set ADC reference and low three bits of analog pin number.
    adcmux[i] = (pin & 7) | ADC_REF;
    if (RECORD_EIGHT_BITS) {
      adcmux[i] |= 1 << ADLAR;
    }

    // If this is the first pin, trigger on timer/counter 1 compare match B.
    adcsrb[i] = i == 0 ? (1 << ADTS2) | (1 << ADTS0) : 0;
#ifdef MUX5
    if (pin > 7) {
      adcsrb[i] |= (1 << MUX5);
    }
#endif  // MUX5
    adcsra[i] = (1 << ADEN) | (1 << ADIE) | adps;
    adcsra[i] |= i == 0 ? 1 << ADATE : 1 << ADSC;
  }

  // Setup timer1
  TCCR1A = 0;
  uint8_t tshift;
  if (ticks < 0X10000) {
    // no prescale, CTC mode
    TCCR1B = (1 << WGM13) | (1 << WGM12) | (1 << CS10);
    tshift = 0;
  } else if (ticks < 0X10000 * 8) {
    // prescale 8, CTC mode
    TCCR1B = (1 << WGM13) | (1 << WGM12) | (1 << CS11);
    tshift = 3;
  } else if (ticks < 0X10000 * 64) {
    // prescale 64, CTC mode
    TCCR1B = (1 << WGM13) | (1 << WGM12) | (1 << CS11) | (1 << CS10);
    tshift = 6;
  } else if (ticks < 0X10000 * 256) {
    // prescale 256, CTC mode
    TCCR1B = (1 << WGM13) | (1 << WGM12) | (1 << CS12);
    tshift = 8;
  } else if (ticks < 0X10000 * 1024) {
    // prescale 1024, CTC mode
    TCCR1B = (1 << WGM13) | (1 << WGM12) | (1 << CS12) | (1 << CS10);
    tshift = 10;
  } else {
    error("Sample Rate Too Slow");
    while (1) {
      ; // stop here
    }
  }
  // divide by prescaler
  ticks >>= tshift;
  // set TOP for timer reset
  ICR1 = ticks - 1;
  // compare for ADC start
  OCR1B = 0;

  // multiply by prescaler
  ticks <<= tshift;
  // Sample interval in CPU clock ticks.
  meta->sampleInterval = ticks;
  meta->cpuFrequency = F_CPU;

  outputMetaDataToSerial(meta);
}

void outputMetaDataToSerial(metadata_t* meta) {
  float sampleRate = (float)meta->cpuFrequency / meta->sampleInterval;
  Serial.print(F("Sample pins:"));
  for (uint8_t i = 0; i < meta->pinCount; i++) {
    Serial.print(' ');
    Serial.print(meta->pinNumber[i], DEC);
  }
  Serial.println();
  Serial.print(F("ADC bits: "));
  Serial.println(meta->recordEightBits ? 8 : 10);
  Serial.print(F("ADC clock kHz: "));
  Serial.println(meta->adcFrequency / 1000);
  Serial.print(F("Sample Rate: "));
  Serial.println(sampleRate);
  Serial.print(F("Sample interval usec: "));
  Serial.println(1000000.0 / sampleRate, 4);
}

//------------------------------------------------------------------------------
// enable ADC and timer1 interrupts
void adcStart() {
  // initialize ISR
  isrBufNeeded = true;
  isrOver = 0;
  adcindex = 1;

  // Clear any pending interrupt.
  ADCSRA |= 1 << ADIF;

  // Setup for first pin.
  ADMUX = adcmux[0];
  ADCSRB = adcsrb[0];
  ADCSRA = adcsra[0];

  // Enable timer1 interrupts.
  timerError = false;
  timerFlag = false;
  TCNT1 = 0;
  TIFR1 = 1 << OCF1B;
  TIMSK1 = 1 << OCIE1B;
}
//------------------------------------------------------------------------------
void adcStop() {
  TIMSK1 = 0;
  ADCSRA = 0;
}


//------------------------------------------------------------------------------
// Convert binary file to csv file.
void binaryToCsv() {
  uint8_t lastPct = 0;
  block_t buf;
  metadata_t* pm;
  uint32_t t0 = millis();
  char csvName[13];
  StdioStream csvStream;

  if (!binFile.isOpen()) {
    Serial.println(F("No current binary file"));
    return;
  }
  binFile.rewind();
  // Create a new csv file.
  strcpy(csvName, binName);
  strcpy(&csvName[BASE_NAME_SIZE + 3], "csv");

  if (!csvStream.fopen(csvName, "w")) {
    error("open csvStream failed");  // OK to stop here
    return;
  }
  Serial.println();
  Serial.print(F("Writing: "));
  Serial.print(csvName);
  Serial.println(F(" - type s character to stop"));
  serialClearInput();
  adcInit((metadata_t*)&buf);
  pm = (metadata_t*)&buf;
  csvStream.print(F("Interval,"));
  float intervalMicros = 1.0e6 * pm->sampleInterval / (float)pm->cpuFrequency;
  csvStream.print(intervalMicros, 4);
  csvStream.println(F(",usec"));
  for (uint8_t i = 0; i < pm->pinCount; i++) {
    if (i) {
      csvStream.putc(',');
    }
    csvStream.print(F("pin"));
    csvStream.print(pm->pinNumber[i]);
  }
  csvStream.println();
  unsigned long countTotal = 0;

  uint32_t tPct = millis();
  char c = ' ';
  uint32_t bn = 0;
  while ((c != 's') && binFile.read(&buf, 512) == 512) {
    // delay(1000);
    while (Serial.available()) {
      c = Serial.read();
      if (c == 's') {
        break;
      }
    }
    if (buf.count == 0) {
      break;
    }
    countTotal += (buf.count / PIN_COUNT); // sample count
    if (buf.overrun) {
      csvStream.print(F("Missed Samples : "));
      csvStream.println(buf.overrun);
    }
    for (uint16_t j = 0; j < buf.count; j += PIN_COUNT) {
      for (uint16_t i = 0; i < PIN_COUNT; i++) {
        if (i) {
          csvStream.putc(',');
        }
        csvStream.print(buf.data[i + j]);
      }
      csvStream.println();
    }
    if ((millis() - tPct) > 1000) {
      uint8_t pct = binFile.curPosition() / (binFile.fileSize() / 100);
      if (pct != lastPct) {
        tPct = millis();
        lastPct = pct;
        Serial.print(pct, DEC);
        Serial.print('%');
        Serial.print(' ');
        Serial.print(countTotal); Serial.print(F(" samples "));
        microsToHrMinSec((countTotal / 1000.0)*intervalMicros);
        Serial.println();
      }
    }
    bn++;
    //    if (Serial.available()) {
    //      break;
    //    }
  }
  binFile.close();
  csvStream.fclose();
  if (c == 's') {
    Serial.print(F("CSV creation stopped by user after ")); Serial.print(bn);
    Serial.println(F(" buffers"));
  } else {
    Serial.print(F("Done: "));
    Serial.print(0.001 * (millis() - t0));
    Serial.println(F(" Seconds"));
  }
  serialClearInput();

}
void microsToHrMinSec(unsigned long mS) {
  unsigned long secs = (unsigned long)(mS / 1000);
  unsigned long mins = secs / 60;
  secs = secs - mins * 60;
  unsigned long hrs = mins / 60;
  mins = mins - hrs * 60;
  Serial.print(' ');
  Serial.print(hrs); Serial.print(':');
  if (mins <= 9) {
    Serial.print('0');
  }
  Serial.print(mins); Serial.print(':');
  if (secs <= 9) {
    Serial.print('0');
  }
  Serial.print(secs); Serial.print(':');
}

//------------------------------------------------------------------------------
// read data file and check for overruns
void checkOverrun() {
  uint8_t lastPct = 0;
  metadata_t* pm;
  uint32_t bn = 0;
  uint32_t totalBn = 0;
  uint32_t overruns = 0;
  bool headerPrinted = false;
  block_t buf;
  uint32_t bgnBlock, endBlock;

  Serial.println();
  if (processAllFiles) {
    Serial.print(F("Checking missed samples errors in "));
    Serial.println(F("All files"));
  } else {
    if (!binFile.isOpen()) {
      Serial.println(F("No current binary file"));
      Serial.println(F("Use a cmd to select All binary files "));
      Serial.println(F("OR use o cmd to select a single binary file "));
      return;
    } else {
      Serial.print(F("Checking missed samples errors in "));
      Serial.println(binName);
    }
  }


  Serial.println(F(" type s character to stop"));

  adcInit((metadata_t*)&buf);
  pm = (metadata_t*)&buf;
  Serial.print(F("Interval,"));
  float intervalMicros = 1.0e6 * pm->sampleInterval / (float)pm->cpuFrequency;
  Serial.print(intervalMicros, 4);
  Serial.println(F(",usec"));

  unsigned long countTotal = 0;

  uint32_t tPct = millis();
  char c = ' ';
  int i = 0;
  if (processAllFiles) {
    binFile.close();
    while (i = setNextBinName(i)) {
      if (!binFile.open(binName, O_RDONLY)) {
        Serial.print(binName);
        error(" open failed"); // continue
      } else {
        break;
      }
    }
  }
LOOP: // jmp here is processing all files

  if (!binFile.isOpen()) {
    Serial.println(F("No current binary file"));
    return;
  }
  if (!binFile.contiguousRange(&bgnBlock, &endBlock)) {
    error("contiguousRange failed");
  }  
  binFile.rewind();
  bn = 0;

  Serial.print(F("Processing : ")); Serial.println(binName);
  while ((c != 's') && (binFile.read(&buf, 512) == 512)) {

    while (Serial.available()) {
      c = Serial.read();
      if (c == 's') {
        break;
      }
    }
    if (buf.count == 0) {
      break;
    }
    countTotal += (buf.count / PIN_COUNT); // sample count
    if (buf.overrun) {
      overruns += buf.overrun;
      if (!headerPrinted) {
        Serial.println();
        Serial.println(F("Missed samples:"));
        Serial.println(F("fileBlockNumber,sdBlockNumber,missedSamplesCount"));
        headerPrinted = true;
      }
      Serial.print(bn);
      Serial.print(',');
      Serial.print(bgnBlock + bn);
      Serial.print(',');
      Serial.println(buf.overrun);
    }
    if ((millis() - tPct) > 1000) {
      uint8_t pct = binFile.curPosition() / (binFile.fileSize() / 100);
      if (pct != lastPct) {
        tPct = millis();
        lastPct = pct;
        Serial.print(pct, DEC);
        Serial.print('%');
        Serial.print(' ');
        Serial.print(countTotal); Serial.print(F(" samples "));
        microsToHrMinSec((countTotal / 1000.0)*intervalMicros);
        Serial.println();
      }
    }

    bn++;
    totalBn++;
  }
  if (bn == 0) {
    Serial.println(F(" file empty."));
  }
NEXT_FILE: // jump here on file errors
  if (processAllFiles && (c != 's')) {
    binFile.close();
    while (((i = setNextBinName(i)) != 0)) {
      if (!binFile.open(binName, O_RDONLY)) {
        Serial.print(binName);
        error(" open failed");
        continue; // try next file
      } // else
      goto LOOP;
    }
  }
  if (c == 's') {
    Serial.print(F("Check of missed samples stopped by user after ")); Serial.print(bn);
    Serial.println(F(" buffers"));
  } else {
    if (!headerPrinted) {
      Serial.print(F("No errors found in ")); Serial.print(totalBn);
      Serial.println(F(" buffers"));
    } else {
      Serial.print(F("Found ")); Serial.print(overruns); Serial.println(F(" missed samples"));
    }
  }
  serialClearInput();
  processAllFiles = false;
}
//------------------------------------------------------------------------------
void serialClearInput() {
  do {
    delay(10);
  } while (Serial.read() >= 0);
}
//------------------------------------------------------------------------------
bool serialReadLine(char* str, size_t size) {
  size_t n = 0;
  while (!Serial.available()) {
  }
  while (true) {
    int c = Serial.read();
    if (c < ' ') break;
    str[n++] = c;
    if (n >= size) {
      Serial.println(F("input too long"));
      return false;
    }
    uint32_t m = millis();
    while (!Serial.available() && (millis() - m) < 100) {}
    if (!Serial.available()) break;
  }
  str[n] = 0;
  return true;
}

//============================================
void openBinFile() {
  char name[NAME_DIM];
  serialClearInput();
  binFile.close();
  Serial.print(F("Enter file name >"));
  if (!serialReadLine(name, sizeof(name))) {
    return;
  }
  Serial.println(name);
  if (!sd.exists(name)) {
    // Serial.println(name);
    Serial.println(F("File does not exist"));
    return;
  }
  binFile.close();
  if (!binFile.open(name, O_RDWR)) {
    // Serial.println(name);
    Serial.println(F("open failed"));
    return;
  }
  strcpy(binName, name);
  Serial.println(F("File opened"));
  serialClearInput();
}

//------------------------------------------------------------------------------
// dump data file to Serial
void dumpData() {
  block_t buf;
  if (!binFile.isOpen()) {
    Serial.println(F("No current binary file"));
    return;
  }
  binFile.rewind();
  Serial.println();
  serialClearInput();
  Serial.println(F("Type s character to stop"));
  char c = ' ';
  uint32_t bn = 0;
  while (c != 's' && binFile.read(&buf , 512) == 512) {
    //delay(1000);
    while (Serial.available()) {
      c = Serial.read();
      if (c == 's') {
        break;
      }
    }
    if (buf.count == 0) {
      break;
    }
    if (buf.overrun) {
      Serial.print(F("Missed Samples,"));
      Serial.println(buf.overrun);
    }
    for (uint16_t i = 0; i < buf.count; i++) {
      Serial.print(buf.data[i], DEC);
      if ((i + 1) % PIN_COUNT) {
        Serial.print(',');
      } else {
        Serial.println();
      }
    }
    bn++;
  }
  if (c == 's') {
    Serial.print(F("Dump stopped by user after "));  Serial.print(bn);
    Serial.println(F(" buffers"));
  } else {
    Serial.println(F("Done"));
  }
  serialClearInput();
}

// -----------------------------------------------
// returns false when have reached FILE_COUNT files
// setNextBinName(0); initialize bin name

int setNextBinName(int count) {
  if (count == 0) { // first time
    // initialize binName
    strcpy(binName, FILE_BASE_NAME);
    strcat(binName, FILE_BASE_EXTRA);
    return 1;
  }
  count++;
  if (count > FILE_COUNT) {
    return 0; // finished
  }
  // else set next name
  if (binName[BASE_NAME_SIZE + 1] != '9') {
    binName[BASE_NAME_SIZE + 1]++;
  } else {
    binName[BASE_NAME_SIZE + 1] = '0';
    if (binName[BASE_NAME_SIZE] == '9') {
      error("Can't create file name");
      return 0; // stop here
    }
    binName[BASE_NAME_SIZE]++;
  }
  return count;
}

//-----------------------------------------------------------------------------
// max number of blocks to erase per erase call
uint32_t const ERASE_SIZE = 262144L;

void initializeFiles() {
  FatFile file;
  sd.chdir(true);
  if (file.openNext(FatFile::cwd(), O_RDONLY)) {
    file.close();
    Serial.println(F("SD not empty."));
    return;
  }
  // else
  unsigned long mSstart = millis();
  uint32_t bgnBlock, endBlock;
  int i = 0; // number of files created so far
  while ((i = setNextBinName(i))) { // returns 0 after FILE_COUNT files
    // Create new file.
    if (sd.exists(binName)) {
      Serial.print(binName);
      error("File Already exists.");
      return;
    }

    Serial.print(F("Creating new file : "));   Serial.print(binName); Serial.print(F(" at ")); Serial.print((millis() - mSstart) / 1000); Serial.println(F(" sec"));
    unsigned long createStart = micros();
    binFile.close();
    if (!binFile.createContiguous(binName, 512 * FILE_BLOCK_COUNT)) {
      error("createContiguous failed. File Exists?");
      return; // error OK here
    }
    // Get the address of the file on the SD.
    if (!binFile.contiguousRange(&bgnBlock, &endBlock)) {
      error("contiguousRange failed");
      return; // error OK here
    }

    // Flash erase all data in the file.
    //Serial.println(F("Erasing all data"));
    uint32_t bgnErase = bgnBlock;
    uint32_t endErase;
    while (bgnErase < endBlock) {
      endErase = bgnErase + ERASE_SIZE;
      if (endErase > endBlock) {
        endErase = endBlock;
      }
      if (!sd.card()->erase(bgnErase, endErase)) {
        error("erase failed");
        return; // error OK here

      }
      bgnErase = endErase + 1;
    }
    Serial.print(F("CreateFile and erase time (mS):")); Serial.println((micros() - createStart) / 1000);

    binFile.close();
  }
  // reset binName to start
  setNextBinName(0);
  serialClearInput();
}

void showMetaData() {
  block_t block[1];
  // check if already have meta file
  char metaName[NAME_DIM];
  SdBaseFile metaFile;
  strcpy(metaName, FILE_BASE_NAME);
  strcat(metaName, FILE_BASE_META_EXTRA);

  if (!sd.exists(metaName)) {
    error("Log Data has NOT been run yet.");
    Serial.println(F(" Current meta data is :"));
    adcInit((metadata_t*) &block[0]); // meta data not used write to another file
    return;
  }

  // create meta file
  if (!metaFile.open(metaName, O_RDONLY)) {
    Serial.println(metaName);
    error(" open failed");
    return;
  }
  metaFile.rewind();
  if (metaFile.read(&block[0] , 512) != 512) {
    error("Read metadata failed");
  }
  metaFile.close();
  outputMetaDataToSerial((metadata_t*) &block[0]);
  setNextBinName(0);
}

//------------------------------------------------------------------------------
// log data
void logData() {
  uint32_t bgnBlock, endBlock, noBlocks;
  lastOverruns = 0;
  digitalWrite(ERROR_LED_PIN, LOW); // clear overrun light
  binFile.close();

  setNextBinName(0);
  if (!binFile.open(binName, O_RDWR)) {
    Serial.println(binName);
    error(" open failed.  Not intialized yet?");
    return; // try next one
  }

  // check if already have meta file
  char metaName[NAME_DIM];
  SdBaseFile metaFile;

  strcpy(metaName, FILE_BASE_NAME);
  strcat(metaName, FILE_BASE_META_EXTRA);

  if (sd.exists(metaName)) {
    error("Log Data has already been run.");
    return;
  }

  // create meta file
  if (!metaFile.open(metaName, O_RDWR | O_CREAT)) {
    Serial.println(metaName);
    error(" open failed");
    return;
  }
  // Allocate extra buffer space.
  block_t block[BUFFER_BLOCK_COUNT];

  Serial.println();
  // Initialize ADC and timer1.
  adcInit((metadata_t*) &block[0]); // meta data not used write to another file

  if (sizeof(block_t) != metaFile.write(&block[0], sizeof(block_t))) {
    Serial.println(metaName);
    error(" Error writing meta data.");
    return; // try next one
  }
  metaFile.close();

  // Initialize queues. -------------------------------
  emptyHead = emptyTail = 0;
  fullHead = fullTail = 0;

  // Put rest of buffers in the empty queue.
  for (uint8_t i = 0; i < BUFFER_BLOCK_COUNT; i++) {
    emptyQueue[emptyHead] = &block[i];
    emptyHead = queueNext(emptyHead);
  }
  //  end initialize queues --------------------------

  // Create new file.
  Serial.println(F("Logging - type s character to stop"));
  // Wait for Serial Idle.
  Serial.flush();
  serialClearInput();

  binFile.close(); // close any current file
  uint32_t overruns = 0;
  uint32_t count = 0;
  uint32_t maxLatency = 0;
  uint32_t t0 = millis();
  uint32_t t1 = t0;
  bool stopLogging = false;
  bool showStatus = false;

  int i = 0;
  adcStart(); // start collecting samples via interrupts
  while (i = setNextBinName(i)) {
    binFile.close();
    if (!binFile.open(binName, O_RDWR)) {
      Serial.println(binName);
      error(" open failed");
      continue; // try next one
    }

    uint32_t bn = 0;
    if (!stopLogging) {
      // Get the address of the file on the SD.
      if (!binFile.contiguousRange(&bgnBlock, &endBlock)) {
        error("contiguousRange failed");
        continue; // try next file
      }
      noBlocks = endBlock - bgnBlock + 1;

      // Start a multiple block write.
      if (!sd.card()->writeStart(bgnBlock, noBlocks)) {
        error("writeBegin failed");
        continue; // try next file
      }

      bn = 1; // if stop now will always keep first 512byte block in file
      while (1) {
        if (fullHead != fullTail) {
          // Get address of block to write.
          block_t* pBlock = fullQueue[fullTail];

          if (showStatus) {// && (emptyHead != emptyTail)) {  // have at least one empty but on UNO never seems to be the case
            showStatus = false;
            //Serial.println('l'); // logging
          }
          // Write block to SD.
          uint32_t usec = micros();
          if (!sd.card()->writeData((uint8_t*)pBlock)) {
            error("write data failed");
            continue; // try next file
          }
          usec = micros() - usec;
          t1 = millis();
          if (usec > maxLatency) {
            maxLatency = usec;
          }
          count += pBlock->count;

          // Add overruns and possibly light LED.
          if (pBlock->overrun) {
            overruns += pBlock->overrun;
            if (ERROR_LED_PIN >= 0) {
              digitalWrite(ERROR_LED_PIN, HIGH);
            }
          }
          // Move block to empty queue.
          emptyQueue[emptyHead] = pBlock;
          emptyHead = queueNext(emptyHead);
          fullTail = queueNext(fullTail);
          bn++;
          if (bn == noBlocks) {
            // adcStop();
            break;
          }
        }
        if (timerError) {
          error("Missed timer event - rate too high");
          // keep going
        }
        if (Serial.available()) {
          char c = Serial.read();
          if (c == 's') {
            // Stop ISR calls.
            adcStop();
            stopLogging = true;
            if (isrBuf != 0 && isrBuf->count >= PIN_COUNT) {
              // Truncate to last complete sample.
              isrBuf->count = PIN_COUNT * (isrBuf->count / PIN_COUNT);
              // Put buffer in full queue.
              fullQueue[fullHead] = isrBuf;
              fullHead = queueNext(fullHead);
              isrBuf = 0;
            }
          } else if (c == '?') {
            showStatus = true;
          }
        }
        if ((stopLogging) && (fullHead == fullTail)) { // have written the last ones
          break;
        }
      }
      if (!sd.card()->writeStop()) {
        error("writeStop failed");
        // keep going
      }
    }
    // Truncate file if recording stopped early.
    // will truncate all unused files
    if (bn != noBlocks) {
      Serial.print(F("Truncating file : ")); Serial.println(binName);
      if (!binFile.truncate(512L * bn)) {
        error("Can't truncate file");
        // keep going
      }
    }
    binFile.close();
  }
  adcStop();
  if (stopLogging) {
    Serial.println(F("Logging stopped by user."));
  }
  Serial.print(F("Last File closed: "));
  Serial.println(binName);
  Serial.print(F("Max block write usec: "));
  Serial.println(maxLatency);
  Serial.print(F("Record time sec: "));
  Serial.println(0.001 * (t1 - t0), 3);
  Serial.print(F("Sample count: "));
  Serial.println(count / PIN_COUNT);
  Serial.print(F("Samples/sec: "));
  Serial.println((1000.0 / PIN_COUNT)*count / (t1 - t0));
  Serial.print(F("Missed samples : "));
  Serial.println(overruns);
  lastOverruns = overruns;
  Serial.println(F("Done"));
  serialClearInput();

}

// ----------------------------------------------------------------------------
void printMenu() {
  Serial.println();
  Serial.print(F("Ver:")); Serial.print(F(VERSION));  Serial.println(F("  type:"));
  Serial.println(F("? - current status"));
  Serial.println(F("i - initialize files"));
  Serial.println(F("c - convert file to csv"));
  Serial.println(F("f - current file"));
  Serial.println(F("o - open file"));
  Serial.println(F("a - open all files (for use with e)"));
  Serial.println(F("d - dump data to Serial"));
  Serial.println(F("l - list files"));
  Serial.println(F("e - missed samples error details"));
  Serial.println(F("r - record ADC data"));
  Serial.println(F("m - display Meta data"));
  Serial.print('>');
}

//------------------------------------------------------------------------------
void setup(void) {
  if (ERROR_LED_PIN >= 0) {
    pinMode(ERROR_LED_PIN, OUTPUT);
    digitalWrite(ERROR_LED_PIN, LOW); // gnd
  }
  if (ERROR_LED_GND_PIN >= 0) {
    pinMode(ERROR_LED_GND_PIN, OUTPUT);
    digitalWrite(ERROR_LED_GND_PIN, LOW); // gnd
  }
  Serial.begin(115200); //9600);

  // Read the first sample pin to init the ADC.
  analogRead(PIN_LIST[0]);

  Serial.print(F("FreeStack: "));
  Serial.println(FreeStack());

  // Initialize at the highest speed supported by the board that is
  // not over 50 MHz. Try a lower speed if SPI errors occur.
  if (!sd.begin(SD_CS_PIN, SD_SCK_MHZ(50))) {
    sd.initErrorPrint();
    fatalBlink();
  }
  delay(3000);
  printMenu();
  blinkDelay.start(200);
  //errorBlink = true; testing
}

//---------------------------------------------------
bool errorPinStateLOW = true;

void checkErrorBlink() {
  if (blinkDelay.justFinished()) {
    blinkDelay.restart();
    errorPinStateLOW = !errorPinStateLOW;
    if (errorBlink) {
      if (errorPinStateLOW) {
        digitalWrite(ERROR_LED_PIN, LOW);
      } else {
        digitalWrite(ERROR_LED_PIN, HIGH);
      }
    }
  }
}

// ---------------------------------------------------------------------------
void printCurrentOpenFileName() {
  if (!binFile.isOpen()) {
    Serial.println(F("No current binary file"));
  } else {
    Serial.print(F("Current open file : ")); Serial.println(binName);
  }
  Serial.println();
}
//------------------------------------------------------------------------------

void loop(void) {
  while (!Serial.available()) {
    SysCall::yield();
    checkErrorBlink();
  }
  char c = Serial.read(); // read the first one
  Serial.println(c);
  serialClearInput(); // clear the rest

  //  if (ERROR_LED_PIN >= 0) {
  //    digitalWrite(ERROR_LED_PIN, LOW);
  //  }
  // Read any Serial data.
  //  do {
  //    delay(10);
  //  } while (Serial.available() && Serial.read() >= 0);
  if (c == '?') {
    Serial.print(F("Missed Samples in last run : ")); Serial.println(lastOverruns);
    printCurrentOpenFileName();
    printMenu();
  } else {
    c = tolower(c);
    if (c == 'c') {
      binaryToCsv();
    } else if (c == 'l') {
      Serial.println(F("ls:"));
      sd.ls(&Serial, LS_DATE | LS_SIZE);
    } else if (c == 'i') {
      initializeFiles();
    } else if (c == 'f') {
      if (processAllFiles) {
        Serial.println(F("All binary files selected."));
      } else if (binFile.isOpen()) {
        Serial.println(binName);
      } else {
        Serial.println(F("No current file."));
      }
    } else if (c == 'o') {
      openBinFile();
      processAllFiles = false;
    } else if (c == 'a') {
      processAllFiles = false;
      Serial.println(F("Opening files:"));
      int  i = 0; // number of files created so far
      while ((i = setNextBinName(i))) { // returns 0 after FILE_COUNT files
        if (sd.exists(binName)) {
          Serial.println(binName);
          processAllFiles = true;
        } else {
          break;
        }
      }
    } else if (c == 'd') {
      dumpData();
    } else if (c == 'e') {
      checkOverrun();
    } else if (c == 'r') {
      logData();
    } else if (c == 'm') {
      showMetaData();
    } else {
      Serial.println(F("Invalid entry"));
    }
    printMenu();
  }
  checkErrorBlink();
}
#else  // __AVR__
#error This program is only for AVR.
#endif  // __AVR__
